/*
 *             foriana
 * 
 *  FOrensic Ram Image ANAlyzer
 *  written in 2007 and 2008 by niekt0
 *  (niekt0@hysteria.sk, https://hysteria.sk/~niekt0)
 *  (please, mail me any discovered bugs/improvements.)
 * 
 * 
 *  version 0.2.0
 * 
 *  newest version at https://hysteria.sk/~niekt0/foriana
 * 
 *  Under GNU GPL.
 *  See http://gnu.org for more information.
 * 
 *  Use at your own risk.
 * 
*/


/* 2do:
 */

/* Limits of this version:
 */


#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>
#include <sys/timeb.h>
#include <errno.h>
#include <libgen.h>

//#include "search.h"

#include <arch/i386/magic.h> // TODO ifndef

#define round(x) ((x)>=0?(int)((x)+0.5):(int)((x)-0.5))
#define _GNU_SOURCE

#define fversion "0.2.0"
#define FORIANA_CACHE_SIZE 10	//how many lines (variables) are stored in "cache" files

// ------

struct my_image{	// information about file, we are adding
	int guest_block_size;	// size of block in analyzed image
	char architecture[32];	// hw architecture of analyzed image
	char operating_system[32];	// operation system if analyzed image
	unsigned long VPT_address;	// position of VPT (from beginig of file)
	unsigned long long ram_size;	  // size of image in kilobytes
	int ts_hl_comm_dst;	  // distance between comm and headlist fields in task_struct
	int md_hl_name_dst;	  // distance between mame and headlist fields in module
	unsigned long ts_one_of_lists; // address of task list_head of one of taks_struct structures
	unsigned long md_one_of_lists; // address of list list_head of one of module structures
	int ts_hl_mm_dst; // =TS_MAGIC_HL_MM_DST; // Distance from "tasks" list_head to "mm" mmstruct...can vary!
	int mm_pgd_dst; // =MM_MAGIC_PGD_DST;  //Distance from beginning of mm_struct to pgd variable...can vary!

	int ts_headlists_number;  // how many headl_lists there are in task_struct before comm
	int md_headlists_number;  // how many headl_lists there are in module before name
	unsigned long ts_headlists[TS_MAGIC_HEADLIST_MAX]; // pointers to headlist of task_struct of one process
	unsigned long md_headlists[MD_MAGIC_HEADLIST_MAX]; // pointers to headlist of module

	// stuff for listing of open files:
	int ts_hl_fs_dst;	//Distance from "tasks" list_head to "files" files_struct in task_struct
	int fs_fdt_dst;		//Distance from beginning of structure to "fdt" fdtable in files_struct
	int fdt_fd_dst;		//Distance from beginning of structure to "fd" **file in fdtable
	int fl_fpath_dst;	//Distance from beginning of structure to "f_path" path in file structure
	int pt_dentry_dst;	//Distance from beginning of structure to "dentry" dentry in path structure
	int dt_name_dst;	//Distance from beginning of structure to "d_name" qstr in dentry structure
	int qs_name_dst;	//Distance from beginning of structure to "name" *char in qstr. note: Should not change, this structure is elemental
	int fs_vfs_dst;		//Distance from beginning of structure to vfsmount in fs_struct.
	unsigned long one_of_paths;		//Pointer to one of path structs

};


struct my_block {		// one Block of data (one virtual page)
	unsigned long status;	// help status
	char block[BLOCK_SIZE];	// raw data
};

/*
 *  Global variables
 */
struct my_image image1;

char helpstatusfile[20]="foriana.cache";
char *statusfile2,*statusfile=helpstatusfile;
FILE *fs;	//cache file

char *inputfile;	//filename
int inputfd,statusfd;	//input & status output
FILE *f;	//separate descriptor for searching in file

struct stat sbuffer;

char *errormsg;

char tmp_small_page[VPT_MAGIC_BS];

int i,j,opt,mode=1,status=0;
unsigned long pointer;

//Forward declarations:
int list_head_size(unsigned long listhead);
int search_init(char *inputfile2,FILE *f2,const char *marker2);
unsigned long search_next();

// Debug function only
int debug(){
	int i;

	printf("--------- Debug info ---------\n");
	printf("guest_block_size: %d\narchitecture: %s\noperating_system: %s\nVPT_address: %lx\nram_size: %llu\nts_hl_comm_dst: %d\nmd_hl_name_dst: %d\nts_one_of_lists: %lx\nts_hl_mm_dst: %d\nmm_pgd_dst: %d\n",
	image1.guest_block_size,	// size of block in analyzed image
	image1.architecture,	// hw architecture of analyzed image
	image1.operating_system,	// operation system if analyzed image
	image1.VPT_address,	// position of VPT (from beginig of file)
	image1.ram_size,	  // size of image in kilobytes
	image1.ts_hl_comm_dst,	  // distance between comm and headlist 
	image1.md_hl_name_dst,
	image1.ts_one_of_lists, // address of tasks list_head of one of taks_struct structures
	image1.ts_hl_mm_dst, // =TS_MAGIC_HL_MM_DST; // Distance from "tasks" list_head to "mm" mmstruct...can vary!
	image1.mm_pgd_dst); // =MM_MAGIC_PGD_DST;  //Distance from beginning of mm_struct to pgd variable...can vary!
	
	printf("ts_headlists_nuber:%d\n",image1.ts_headlists_number);
	for (i=TS_MAGIC_HEADLIST_MAX;i>=0;i--){
		printf("  Headlist number %d: %lx\n",i,image1.ts_headlists[i]);
	}
	
	printf("md_headlists_nuber:%d\n",image1.md_headlists_number);
	for (i=MD_MAGIC_HEADLIST_MAX;i>=0;i--){
		printf("  Headlist number %d: %lx\n",i,image1.md_headlists[i]);
	}

	printf("---------------------------------\n");

	return 0;
}

int save_info(){
	int i;
	
	printf("save_info: \n");

        if ( (fs = fopen(statusfile,"w")) == NULL ) {
		errormsg = strerror(errno);
		fprintf(stderr, "%s: %s.\n",statusfile,errormsg);
		return 22;
	}

	// So at this moment we have file open for writing

	
	i=fprintf(fs,"%d\n%s\n%s\n%lx\n%llu\n%d\n%d\n%lx\n%d\n%d\n",
	image1.guest_block_size,	// size of block in analyzed image
	image1.architecture,	// hw architecture of analyzed image
	image1.operating_system,	// operation system if analyzed image
	image1.VPT_address,	// position of VPT (from beginig of file)
	image1.ram_size,	  // size of image in kilobytes
	image1.ts_hl_comm_dst,	  // distance between comm and headlist 
	image1.md_hl_name_dst,
	image1.ts_one_of_lists, // address of tasks list_head of one of taks_struct structures
	image1.ts_hl_mm_dst, // =TS_MAGIC_HL_MM_DST; // Distance from "tasks" list_head to "mm" mmstruct...can vary!
	image1.mm_pgd_dst); // =MM_MAGIC_PGD_DST;  //Distance from beginning of mm_struct to pgd variable...can vary!
	
	i=fclose(fs);
	
	return i;
}

int read_info(){
	int i;

	printf("read_info: %s\n",statusfile);fflush(stdout);

        if ( (fs = fopen(statusfile,"r")) == NULL ) {
		printf("read_info: file does not exist.\n");fflush(stdout);
		return 1;	//file does not exists => we will perform determination
	}

	// So at this moment we have file open for reading

	
	i=fscanf(fs,"%d\n%s\n%s\n%lx\n%llu\n%d\n%d\n%lx\n%d\n%d\n",
	&image1.guest_block_size,	// size of block in analyzed image
	image1.architecture,	// hw architecture of analyzed image
	image1.operating_system,	// operation system if analyzed image
	&image1.VPT_address,	// position of VPT (from beginig of file)
	&image1.ram_size,	  // size of image in kilobytes
	&image1.ts_hl_comm_dst,	  // distance between comm and headlist 
	&image1.md_hl_name_dst,
	&image1.ts_one_of_lists, // address of tasks list_head of one of taks_struct structures
	&image1.ts_hl_mm_dst, // =TS_MAGIC_HL_MM_DST; // Distance from "tasks" list_head to "mm" mmstruct...can vary!
	&image1.mm_pgd_dst); // =MM_MAGIC_PGD_DST;  //Distance from beginning of mm_struct to pgd variable...can vary!
	
//	printf("read_info: elements: %d.\n",i);fflush(stdout);
	
	if (i != FORIANA_CACHE_SIZE){
		printf("read_info: %d elements: bad file format.\n",i);fflush(stdout);
		return 2;
	}	

	return 0;
}


/*
 * Some alchemy:
 */

/*
 * Function that manages mapping from linear to physical address space.
 * 
 */
int read_linear(unsigned long PGD, unsigned long lin_addr, void *out_data, unsigned int priv_block_size){

	//"PGD" is needed, because each process have it's own linear address space. Process 0 means kernel memory.

	//Find in VPT "physical" address, where data should be. (if problem, return 1)
	//Read block_size of that data into out_data (if problem, return 2)
	//return 0

	//unsigned long PDE_addr; //= find_PDE();

	unsigned long physical_offset = lin_addr;
	unsigned long PDE_offset = lin_addr;
	unsigned long PTE_offset = lin_addr;
	unsigned long page_number = lin_addr;
	unsigned long tmp_var = 0;

	//---
	
//	 printf("read_linear on addr: 0x%X,\n",lin_addr); //DBG

	if (PGD == 0) {	//special case
		PGD = image1.VPT_address;
	}//else {	//PGD is linear address of PGD of some process
		
//		PGD=read_linear(0,PGD,&PGD,POI);
//	}

	
	//Kernel linear address space:

	// For extracting addresses I use bitwise and, for extracting offsets I have to shift bits.
        physical_offset = physical_offset & 0x003FFFFF;//  << 10; >>10;//bits 21:0  // 4-Mbyte Page Translation
        PDE_offset = PDE_offset >> 22; //bits 31:22 

//	printf("read_linear: %lx, addPO: 0x%X, PDE_o: 0x%X, PDE_e: 0x%X\n",lin_addr,physical_offset,PDE_offset,image1.VPT_address+PDE_offset*POINTER_SIZE); //DBG

        lseek(inputfd,image1.VPT_address+PDE_offset*POINTER_SIZE,SEEK_SET);// *4 because each entry have 32 bits = 4bytes

        if ( (status=read(inputfd,&page_number,POINTER_SIZE)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
                return 35;
        }

//	printf("page_number: 0x%X, pointer size: %d\n",page_number,POINTER_SIZE); //DBG

	// TODO: process all necessary special bit, like swapped, etc...

	tmp_var = page_number;
	tmp_var = tmp_var & 0x00000080; //only bit NUMBER 7 (ie. 8th bit)

	if (tmp_var != 0) { //4MB pages are used
		
//		printf("4MB pages used for this page.\n"); //DBG

		page_number = page_number >> 22; //bits 31:22

		//If we wan't to read more than 1 page:
		// (use recursion;)
		//
		if (priv_block_size > (MB_4-physical_offset)){ // offset + block_size must be less then 4M 
	//		printf("read_linear: page too small for reading... processing another one.\n");
			//recursion
			
			tmp_var = MB_4-physical_offset;

			read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var);

		} else{
			tmp_var=priv_block_size;
		}	

		lseek(inputfd,page_number*MB_4+physical_offset,SEEK_SET); //(4MB page number "large_page_number" + offset)
//		printf("page_number: 0x%X, phys. addr: 0x%X\n",page_number,page_number*MB_4+physical_offset); //DBG
		
		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
                	fprintf(stderr, "read_linear: read() in %s: %s.\n",inputfile,errormsg);
	                return 35;
        	}
	} else {   //4KB pages are used
//		printf("4KB pages used for this page.\n"); //DBG
	
		page_number = page_number & 0xFFFFF000; //bits 31:12 ... This must be done this way, not like commented out example

		PTE_offset = (lin_addr & 0x003FF000) >> 12; //bits 21:12 ... This is offset, so we have to shift bits.//XXX optimalize?
		
		physical_offset = physical_offset & 0x00000FFF;//  << 10; >>10;//bits 11:0
		
	//	page_number = page_number >> 12; //bits 31:12 //PDE entry ??? WTF

//		printf("Searching for PTE in 0x%lx.\n",page_number); //DBG

		lseek(inputfd,page_number+PTE_offset*4,SEEK_SET); // *4 because each entry have 32 bits = 4bytes //??? XXX

		if ( (status=read(inputfd,&tmp_var,4)) == -1 ) {
                 	errormsg=strerror(errno);
		        fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
			return 36;
		}
		
//		printf("In PTE I found 0x%lx.\n",tmp_var); //DBG

		//tmp_var = tmp_var >> 12; //bits 31:12  //PTE entry
		tmp_var = tmp_var & 0xFFFFF000; // this or this way?

//		printf("Seeking to address  0x%lx.\n",tmp_var+physical_offset); //DBG

		lseek(inputfd,tmp_var+physical_offset,SEEK_SET);

		//If we want to read more than 1 page:
		// (use recursion;)
		//
		if (priv_block_size > (4096-physical_offset)){ // offset + block_size must be less then 4M
	//		printf("read_linear: page too small for reading... processing another one.\n");
			//recursion
			
			tmp_var = 4096-physical_offset; //wtf?? less then zero?? XXX

			read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var); //Process rest of request recursively... beware about effectivity

		} else{
			tmp_var=priv_block_size;
		}

		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
                	fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
	                return 38;
        	}


	}

	return 0;

}

/*
 * Function for verifying if given address is inside task_struct.
 *
 * Verification is done by finding and checking double linked 
 * "list_head" "tasks". This way is chosen instead of classic 
 * pattern matching, because I believe it is more flexible and 
 * can be used for almost every 2.6 kernel version. (even future;)
 * It does not depend on task_struct definition too much, only 
 * important characteristic is number of list_heads between 
 * "comm" array and list_head "tasks". This is specified by
 * TS_MAGIC_HEADLIST_NUMBER. 
 *
 * Correct list_head can be verified easily:
 * Data at address point to (virtual) address where is stored another
 * list_head. Second pointer in this second head list point back
 * to (virtual) address where we started.
 *
 *  ...   |--------> |---------->  |---------->   ...
 *          <------  | <-----------| <-------- 
 *
 * Note: This will not find processes, that are inactive,
 * has already ended in past, or something similiar. Therefore,
 * if you want to mine as much data as possible, and don't 
 * afraid of too much false positives, use different method.
 * (idetect toolkit, or some greping shell utilities)
 *
 *
 *  verify_list_head (address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int verify_list_head(unsigned long ver_addr){

	unsigned long tmp_addr1,tmp_addr2,tmp_addr3;

	if ( (status=lseek(inputfd,ver_addr,SEEK_SET)) == -1 ) {
		errormsg = strerror(errno);
                fprintf(stderr, "verify_list_head: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
                return(1);
        }

	if ( (status=read(inputfd,&tmp_addr1,POINTER_SIZE)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "verify_list_head: read() in %s: %s.\n",inputfile,errormsg);
                return(2);
        }

	if (tmp_addr1 == 0) {
//		printf("verify_list_head: Zero.\n");
		return 6;
	}

//	printf("Verify_list_head: Co tam je:%X\n",tmp_addr1);

	if ( (status = read_linear(0,tmp_addr1,&tmp_addr2,POINTER_SIZE) == -1 )) {
                printf("verify_list_head: 0x%lx seems is not valid virtual address.\n",tmp_addr1);
                return(3);
        }
	
//	printf("Verify_list_head: Co tam je2:%X\n",tmp_addr2);	

	if ( (status = read_linear(0,tmp_addr2+POINTER_SIZE,&tmp_addr3,POINTER_SIZE) == -1 )) {
                printf("verify_list_head: 0x%lx seems is not valid virtual address.\n",tmp_addr3);
                return(4);
        }
	
	if (tmp_addr1 != tmp_addr3) {
//		printf("verify_list_head: addresses 0x%lx and 0x%lx does not match.\n",tmp_addr1,tmp_addr3);
		return(5);
	}
	
//	printf("		Verify list_head: pointer to: %lx\n",tmp_addr2);

	//We need to verify second pointer of list_head too, there were some false positives
	// same algoritmus, except that "prev" pointer points to (addres - POINTER_SIZE) of reverse pointer.
	

	if ( (status=lseek(inputfd,ver_addr-POINTER_SIZE,SEEK_SET)) == -1 ) {
		errormsg = strerror(errno);
                fprintf(stderr, "verify_list_head: 2: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
                return(6);
        }

	if ( (status=read(inputfd,&tmp_addr1,POINTER_SIZE)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "verify_list_head: 2: read() in %s: %s.\n",inputfile,errormsg);
                return(7);
        }

	if (tmp_addr1 == 0) {
//		printf("verify_list_head: Zero.\n");
		return 8;
	}

//	printf("Verify_list_head: 3:Co tam je:%X\n",tmp_addr1);

	if ( (status = read_linear(0,tmp_addr1+POINTER_SIZE,&tmp_addr2,POINTER_SIZE) == -1 )) {
                printf("verify_list_head: 2: 0x%lx seems is not valid virtual address.\n",tmp_addr1);
                return(9);
        }
	
//	printf("Verify_list_head: 3: Co tam je2:%X\n",tmp_addr2);	

	if ( (status = read_linear(0,tmp_addr2,&tmp_addr3,POINTER_SIZE) == -1 )) {
                printf("verify_list_head: 2: 0x%lx seems is not valid virtual address.\n",tmp_addr3);
                return(10);
        }
	
	if (tmp_addr1 != tmp_addr3) {
//		printf("verify_list_head: 2: addresses 0x%lx and 0x%lx does not match.\n",tmp_addr1,tmp_addr3);
		return(11);
	}
	
//	printf("Verify list_head: 2: pointer to: %lx\n",tmp_addr2);

	return 0;
}


/*
 *  verify_vfsmount (pointer to vfsmount structure)
 *  return 0 on success, otherwise -1
 *
 *  vfsmount structure is "recognized" by simple test:
 *  should contain pointer to string "/dev/".
 *  (char *mnt_name)
 *  This is alone not a strong heuristic, but in combination
 *  with specific chain where structure should be found when trying 
 *  to list open files, it is enough.
 */

int verify_vfsmount(unsigned long ver_addr) {

	int tmp,tmp2;
	unsigned long tmp3;
	char pattern[VFS_MAGIC_PATTERN_LEN];

	for (tmp=1;tmp<VFS_MAGIC_MAX_TRIES;tmp++){
	
		printf("	verify_vfsmount: %d: %lx ::\n",tmp,ver_addr+(tmp*POINTER_SIZE)); //DBG

		if ( (status = read_linear(0,ver_addr+(tmp*POINTER_SIZE),&tmp3,POINTER_SIZE) == -1 )) { //read pointer to string
                	printf("	verify_vfsmount: 0x%lx seems is not valid virtual address.\n",ver_addr+(tmp*POINTER_SIZE));
                //	return -3;
       		 } else {

			if ( (status = read_linear(0,tmp3,pattern,VFS_MAGIC_PATTERN_LEN) == -1 )) { //read string
	      	        	printf("	verify_vfsmount: 0x%lx seems is not valid virtual address.\n",tmp3);
	        	  //      return -4;
	        	} else {
	
				tmp2=strncmp(pattern,VFS_MAGIC_PATTERN,VFS_MAGIC_PATTERN_LEN); //compare found string
				
				if (tmp2 == 0) {	// found "/dev/"
					printf("	verify_vfsmount: Found some \"/dev/\" at offset %d\n",tmp*POINTER_SIZE); //DBG
						return 0;  // Found => It is task_list
					}		
				}
			}
		}

	return -1;	// characterstic string not found => prob. not vfsmount
}



/*
 *  verify_qstr (pointer to "qstr" structure )
 *  return 0 on success, otherwise error
 *
 *  Used for listing of open files per process.
 *    d_name -> name
 *    (qstr)   (char)
 *
 *  This function verifies, if pointer points to qstr structure.
 *  qstr is elemental structure, and we assume, that will not change (there is no reason)
 *  Verification is done by checking, if string that "name" points to is really "len"
 *  bytes long (looking for \0) and "len" is not 0.
 *
 *  This is qstr definition:
 *
 *  struct qstr {
 *          unsigned int hash;
 *          unsigned int len;
 *          const unsigned char *name;
 *  };
 *                          
 *
 */

int verify_qstr(unsigned long ver_addr) {
	
	unsigned long tmp1,tmp2;
//	size_t tmp3;
	char tmp_filename[QS_MAGIC_MAX_FILENAME_SIZE];

	if ( (status = read_linear(0,ver_addr+INTEGER_SIZE,&tmp1,INTEGER_SIZE) == -1 )) { //skip "hash", read string length
                printf("verify_qstr: 0x%lx seems is not valid virtual address.\n",ver_addr+INTEGER_SIZE);
                return(2);
        }

	if ( (status = read_linear(0,ver_addr+(INTEGER_SIZE*2),&tmp2,POINTER_SIZE) == -1 )) { //read pointer to string
                printf("verify_qstr: 0x%lx seems is not valid virtual address.\n",ver_addr+INTEGER_SIZE);
                return(3);
        }

	if (tmp1 > QS_MAGIC_MAX_FILENAME_SIZE) {
		printf("verify_qstr: length too big: %lx.\n",tmp1);
		return -2;
	}

	if ( (status = read_linear(0,tmp2,tmp_filename,tmp1) == -1 )) { //read pointer to string
                printf("verify_qstr: 0x%lx seems is not valid virtual address.\n",ver_addr+INTEGER_SIZE);
                return(4);
        }

//	tmp3=tmp1+1;	//TODO XXX find out why this gives:
//	warning: implicit declaration of function ‘strnlen’
//	tmp3=strnlen((const char *)tmp_filename,tmp3);

	if ( strnlen(tmp_filename,tmp1+1) != (tmp1)) {
		printf("verify_qstr: sizes does not match: %lx vs %d.\n",tmp1,strlen(tmp_filename));
		return -1;
	}


	//try to read whole string
	//if size is ok, return 0

	return 0;
}

/*
 *  verify_dentry (pointer to "dentry" structure )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into image1
 *
 *  Used for listing of open files per process.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  This function determines distances in structure dentry
 *  It is better to determine structures by more "sub" heuristics, because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int verify_dentry(unsigned long ver_addr) {

	int tmp,tmp2;

	for (tmp=1;tmp<DT_MAGIC_MAX_TRIES;tmp++){
	
		printf("verify_dentry: %d: %lx ::\n",tmp,ver_addr+(tmp*POINTER_SIZE)); //DBG
		tmp2=verify_qstr(ver_addr+(tmp*POINTER_SIZE));
		
		if (tmp2 == 0) {	// found qstr
		
			printf("	verify_dentry: Found some qstr at offset %d\n",tmp*POINTER_SIZE); //DBG

			image1.dt_name_dst=tmp*POINTER_SIZE; //save determined distance

				return 0;  // Found => It is task_list
			}		
		}

	return -1;	// qstr not found
}

/*
 *  determine_up_to_path (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into image1
 *
 *  function that tries to determine distances in structures needed
 *  for listing of open files per process up to f_path structure.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  So this function determines distances in structures files_struct, fdtable, file and path.
 *  It is better to determine structures by more "sub" heuristics, because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int determine_up_to_path(unsigned long ver_addr) {

	int tmp,tmp2,tmp3,tmp4;
	unsigned long tmpl1,tmpl2,tmpl3,tmpl4;

	for (tmp=1;tmp<TS_MAGIC_FILES_MAX_TRIES;tmp++){	//---Looking for "files" structure
		printf("	determine_up_to_path1: %d: %lx ::\n",tmp,ver_addr+(tmp*POINTER_SIZE)); //DBG
		if ( (status = read_linear(0,ver_addr+(tmp*POINTER_SIZE),&tmpl1,POINTER_SIZE) == -1 )) { //read pointer to "files"
                	printf("	determine_up_to_path: 0x%lx seems is not valid virtual address.\n",ver_addr+(tmp*POINTER_SIZE));
       		 } else {

			for (tmp2=1;tmp2<FS_MAGIC_MAX_TRIES;tmp2++){	//---Looking for "fdt" structure
				printf("	determine_up_to_path2: %d: %lx ::\n",tmp2,tmpl1+(tmp2*POINTER_SIZE)); //DBG
				if ( (status = read_linear(0,tmpl1+(tmp2*POINTER_SIZE),&tmpl2,POINTER_SIZE) == -1 )) { //read pointer to string
		                	printf("	determine_up_to_path: 0x%lx seems is not valid virtual address.\n",tmpl1+(tmp2*POINTER_SIZE));
		       		 } else {

					for (tmp3=1;tmp3<FD_MAGIC_MAX_TRIES;tmp3++){	//---Looking for "fd" structure
						printf("	determine_up_to_path4: %d: %lx ::\n",tmp3,tmpl2+(tmp3*POINTER_SIZE)); //DBG
						if ( (status = read_linear(0,tmpl2+(tmp3*POINTER_SIZE),&tmpl3,POINTER_SIZE) == -1 )) { //read pointer to string
				                	printf("	determine_up_to_path: 0x%lx seems is not valid virtual address.\n",tmpl2+(tmp3*POINTER_SIZE));
				       		 } else {

							for (tmp4=1;tmp4<FL_MAGIC_MAX_TRIES;tmp4++){	//---Looking for "fd" structure
								printf("	determine_up_to_path5: %d: %lx ::\n",tmp4,tmpl3+(tmp4*POINTER_SIZE)); //DBG
								if ( (status = read_linear(0,tmpl3+(tmp4*POINTER_SIZE),&tmpl4,POINTER_SIZE) == -1 )) { //read pointer to string
						                	printf("	determine_up_to_path: 0x%lx seems is not valid virtual address.\n",tmpl3+(tmp4*POINTER_SIZE));
						       		 } else {
			
									tmp4=verify_vfsmount(tmpl4); // Now call verification function	
									if (tmp4 == 0) {	// found !
										printf("	determine_up_to_path6: Found some vfs_struct at offset %lx\n",tmpl3); //DBG
										image1.ts_hl_fs_dst=tmp*POINTER_SIZE;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.fdt_fd_dst=tmp2*POINTER_SIZE;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.fl_fpath_dst=tmp3*POINTER_SIZE;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.fs_vfs_dst=tmp4*POINTER_SIZE;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.one_of_paths=tmpl3;		//Linear pointer to one of path structures

										return 0;  // Found => It is task_list
									}
								}	
							}	
						}	
					}		
				}
			}
		}

	}

	return -1;	// Did not found
}

/*
 *  determine_open_files (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *
 *  function that tries to determine distances in all structures needed
 *  for listing of open files per process. 
 *  This will definitely not work in 2.4 kernels.
 */

int determine_open_files(unsigned long ver_addr) {

	int tmp,tmp2;
	
	printf(" determine_open_files: 0x%lx.\n",ver_addr);
	tmp=determine_up_to_path(ver_addr+image1.ts_hl_comm_dst);	//"files" is after "comm" field

	if (tmp != 0) {
		printf(" determine_open_files: determine_up_to_path  on 0x%lx returned  %d.\n",ver_addr+image1.ts_hl_comm_dst,tmp);
		return -2;
	}
	for (tmp=1;tmp<FL_MAGIC_MAX_TRIES;tmp++){	//---Looking for "fd" structure
		printf("	determine_open_files2: %d: %lx ::\n",tmp,image1.one_of_paths+(tmp*POINTER_SIZE)); //DBG
		tmp2=verify_dentry(image1.one_of_paths+(tmp*POINTER_SIZE)); // Now call verification function	
	
		if (tmp2 == 0) {	// found !
			printf("	determine_open_files3: Found some dentry at offset %lx\n",image1.one_of_paths+(tmp*POINTER_SIZE)); //DBG
			image1.pt_dentry_dst=tmp*POINTER_SIZE;	//save offset inside path structure (to dentry)
			printf("Determination for listing of open files completed.\n");			
			return 0;  // Found => It is task_list	
		}
	}
		
	return -1;
}

/*
 *  verify_task_struct (pointer to "comm" field of task_struct )
 *  return 0 on success, otherwise -1
 */

int verify_task_struct(unsigned long ver_addr) {

	int tmp,tmp2,tmp3=TS_MAGIC_HEADLIST_MAX,tmp5=-1,tmp6;
	unsigned long tmp4;

	for (tmp=1;tmp<TS_MAGIC_MAX_TRIES;tmp++){
	
//		printf("verify_task_struct: %d: %lx ::\n",tmp,ver_addr-(tmp*POINTER_SIZE)); //DBG
		tmp2=verify_list_head(ver_addr-(tmp*POINTER_SIZE));
		
		if (tmp2 == 0) {	// found some list_head
		
//			printf("	verify_task_struct: Found some list_head at negative offset %d\n",tmp*POINTER_SIZE); //DBG

			image1.ts_headlists[++tmp5]=ver_addr-(tmp*POINTER_SIZE); //save each list_head
//			printf("	verify_task_struct: headlists[%d]:%lx \n",tmp5,image1.ts_headlists[tmp5]);

			tmp++; /* because of there can be (and are) 3 head lists in row, and without this
				  our algorithm would find 5 "list_heads" */

			if (--tmp3 == 0 ) { //maximum list_head's found
					    // this should not happen
				return -7;  // Found => It is task_list
			}		
		}
	}

//	printf("        verify_task_struct: tmp5: %d\n",tmp5);
	
	if (tmp5 >= TS_MAGIC_HEADLIST_NUMBER) { //If this is really task_struct

		tmp2=tmp5-TS_MAGIC_HEADLIST_FRONT_NUMBER;

		//This step is needed, because verify_list_head seems to produce 
		// some false positives, so TS_MAGIC_HEADLIST_FRONT_NUMBER may be not enough accurate.
		tmp3=0; tmp6=0;
		for (tmp=tmp2-TS_MAGIC_HEADLIST_SPREAD;tmp < (tmp2+TS_MAGIC_HEADLIST_SPREAD);tmp++){ 

			tmp3=list_head_size(image1.ts_headlists[tmp]);

			if (tmp3 > tmp6) {// more and not loop (loop < 0)
				tmp6=tmp3;			//remember local maximum
				image1.ts_headlists_number=tmp; //save which list_head is right one
			}
		}

		// So at this point we know position of right list_head

//		printf("	verify_task_struct: Found tasks at negative offset %d\n",tmp*POINTER_SIZE); //DBG
		image1.ts_hl_comm_dst=ver_addr-image1.ts_headlists[image1.ts_headlists_number]+POINTER_SIZE; //set distance from headlist to comm for this kernel.
		
		//XXX I change this to be linear address XXX now it points to another process (2 right) is it prob.?
		if ( (status=lseek(inputfd,image1.ts_headlists[image1.ts_headlists_number],SEEK_SET)) == -1 ) {
	                errormsg = strerror(errno);
	                fprintf(stderr, "verify_task_struct: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
	                return(-2);
	        }

	        if ( (status=read(inputfd,&tmp4,POINTER_SIZE)) == -1 ) {
	                errormsg=strerror(errno);
	                fprintf(stderr, "verify_list_head: read() in %s: %s.\n",inputfile,errormsg);
	                return(-3);
	        }
	
		image1.ts_one_of_lists=tmp4;
		return 0;  // Found => It is task_list
	}
	//otherwise it is not task_struct
	return -1;
}

/*
 *
 *  Interesting fields in "module" are just near. This should not change (no reason),
 *  so this function is much simpler then verify_task_struct .
 *
 *  verify_module (pointer to "name" field of module )
 *  return distance between name and list_head in bytes on success, otherwise < 0
 *
 *  
 */

int verify_module(unsigned long ver_addr) {

	int tmp,tmp2,tmp3=MD_MAGIC_HEADLIST_MAX,tmp5=-1,tmp6;
	unsigned long tmp4;

	for (tmp=1;tmp<MD_MAGIC_MAX_TRIES;tmp++){
	
//		printf("verify_module: %d: %lx ::\n",tmp,ver_addr-(tmp*POINTER_SIZE)); //DBG
		tmp2=verify_list_head(ver_addr-(tmp*POINTER_SIZE)); /// XXX - and +
		
		if (tmp2 == 0) {	// found some list_head
		
			printf("	verify_module: Found some list_head at negative offset %d\n",tmp*POINTER_SIZE); //DBG

			image1.md_headlists[++tmp5]=ver_addr-(tmp*POINTER_SIZE); //save each list_head
			printf("	verify_module: headlists[%d]:%lx \n",tmp5,image1.ts_headlists[tmp5]);

			tmp++; /* because of there can be (and are) 3 head lists in row, and without this
				  our algorithm would find 5 "list_heads" */

			if (--tmp3 == 0 ) { //maximum list_head's found
					    // this should not happen
				printf("	verify_module: max reached. Not ideal.\n");
				break; //don't give up
//				return -7;  // Found => It is task_list
			}		
		}
	}

	printf("        verify_module: tmp5: %d\n",tmp5);
	
	if (tmp5 >= 1) { //If this is really task_struct

		//This step is needed, because verify_list_head seems to produce 
		// some false positives, so TS_MAGIC_HEADLIST_FRONT_NUMBER may be not enough accurate.
		tmp3=0; tmp6=0;
		for (tmp=0;tmp <= tmp5;tmp++){ //or tmp5+1?

			tmp3=list_head_size(image1.ts_headlists[tmp]);
			printf("        verify_module: tmp3 (size): %d of %lx\n",tmp3,image1.md_headlists[tmp]);
			if (tmp3 > tmp6) {// more and not loop (loop < 0)
				tmp6=tmp3;			//remember local maximum
				image1.md_headlists_number=tmp; //save which list_head is right one
			}
		}

		// So at this point we know position of right list_head

		printf("	verify_module: Found tasks at negative offset %d\n",tmp*POINTER_SIZE); //DBG
		image1.md_hl_name_dst=ver_addr-image1.md_headlists[image1.md_headlists_number]+POINTER_SIZE; //set distance from headlist to name for this kernel.
		
		//XXX I change this to be linear address XXX now it points to another process (2 right) is it prob.?
		if ( (status=lseek(inputfd,image1.md_headlists[image1.md_headlists_number],SEEK_SET)) == -1 ) {
	                errormsg = strerror(errno);
	                fprintf(stderr, "verify_module: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
	                return(-2);
	        }

	        if ( (status=read(inputfd,&tmp4,POINTER_SIZE)) == -1 ) {
	                errormsg=strerror(errno);
	                fprintf(stderr, "verify_module: read() in %s: %s.\n",inputfile,errormsg);
	                return(-3);
	        }
	
		image1.md_one_of_lists=tmp4;
		return 0;  // Found => It is task_list
	}
	//otherwise it is not task_struct
	return -1;
}

/*
int verify_module(unsigned long ver_addr) {

	int tmp,tmp2,tmp3=TS_MAGIC_HEADLIST_MAX,tmp5=-1,tmp6;
	unsigned long tmp4;

	tmp2=verify_list_head(ver_addr-POINTER_SIZE);	//This should not change.... ok for now
	
	if (tmp2 == 0) {	// found some list_head
		
		image1.md_one_of_lists=ver_addr-POINTER_SIZE;
		printf("	verify_module: Found some list_head. %d\n"); //DBG

	} else {	
		return -1;	// Not module
	}

	//XXX I change this to be linear address XXX now it points to another process (2 right) is it prob.?
	if ( (status=lseek(inputfd,image1.md_one_of_lists,SEEK_SET)) == -1 ) {
                errormsg = strerror(errno);
                fprintf(stderr, "verify_task_struct: lseek() to 0x%lx failed because of %s.\n",ver_addr,errormsg);
                return(-2);
        }

        if ( (status=read(inputfd,&tmp4,POINTER_SIZE)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "verify_list_head: read() in %s: %s.\n",inputfile,errormsg);
                return(-3);
        }

	image1.md_one_of_lists=tmp4;
	return 0;  // Found => It is module
}
*/

/* 
 * Initialization function
 *
 * return > 0 if error occure
 * otherwise 0
 */

int init(){

//	printf("Init,%d, %s\n",inputfd,inputfile);

	//open input file	
        if ( (inputfd = open(inputfile,O_RDONLY)) == -1 ) {
                errormsg = strerror(errno);
                fprintf(stderr, "%s: %s.\n",inputfile,errormsg);
                return 21;
        }
	
//	printf("Init2:,%d\n",inputfd);

	//Determine file size
	if ( fstat(inputfd,&sbuffer) == -1){
		fprintf(stderr, "%s: Cannot get file statistics.\n",inputfile);
		return 22;
	}

	image1.ram_size=sbuffer.st_size;

	strcpy(image1.architecture,"NULL");
	strcpy(image1.operating_system,"NULL");

	//Check if we can read image "header"
	if (image1.ram_size < MB_1){	//XXX 1MB 
		fprintf(stderr, "%s: File is too small (<1MB). Are you sure it is a memmory dump? If yes, please modify foriana source code.\n",inputfile);
		return 23;
	}
	
	printf("Dumped RAM size is %llu MB.\n",image1.ram_size/MB_1); //strange results...
	
	return 0;

}

/*
 * Function for verifying if given address contains VPT 
 * (resp. PaGeDirectory) of some process.
 *
 * Verification is done by simple checking
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, as userspace PGD are 
 * cloned from kernel PGD, but it is just heuristic, 
 * and can be wrong.
 *
 *  user_VPT_test (linear address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */


int user_VPT_test(unsigned long my_addr){

	int mi=0;

	if ( (status=read_linear(0,my_addr,tmp_small_page,VPT_MAGIC_BS)) != 0 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "user_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
//	printf("user_VPT_test: %lx\n",my_addr);

	mi=VPT_MAGIC_ZEROES; 

	//chceck if structure seems to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	//see documentation for more information

//	printf("dbg: %2X %2X %2X\n",tmp_small_page[mi]-0xFFFFFF00,tmp_small_page[mi+4]-0xFFFFFF00,tmp_small_page[mi+8]-0xFFFFFF00); 
//	printf("user_vpt_test: dbg :i: %d: %X %X %X\n",mi,tmp_small_page[mi],tmp_small_page[mi+4],tmp_small_page[mi+8]); 

	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+POINTER_SIZE] & 0x6f)) {return 2;}
	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+POINTER_SIZE*2] & 0x6f)) {return 3;}
	if (tmp_small_page[mi] == 0) {return 4;}
	if (tmp_small_page[mi] == 0xFFFFFFFF) {return 5;}
//	if ((tmp_small_page[mi] & 0x000000FF) != 0x000000E3) {return 6;} //XXX Just try XXX

//	printf("user_vpt_test: dbg2: :%X,%d,%c\n",tmp_small_page[mi]),tmp_small_page[mi],tmp_small_page[mi];

	return 0; //all OK			
}

/*
 * Function for verifying if given address contains system VPT 
 * (resp. PaGeDirectory)
 *
 * Verification is done by simple checking
 * if first VPT_MAGIC_ZEROES bytes contain only zeroes and 
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, but it is just heuristic, 
 * and can be wrong.
 *
 *  VPT_test (physical address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int VPT_test(int my_addr){

	int mi=0;

	if ( (status=read(inputfd,&tmp_small_page,VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < VPT_MAGIC_ZEROES;mi++) { //first 
		if (tmp_small_page[mi] != 0) {
			return 1;
		}
	}
	
	//check if structure seems to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	//see documentation for more information

	//0x6f ~ 11011111 
	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+POINTER_SIZE] & 0x6f)) {return 2;}
	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+POINTER_SIZE*2] & 0x6f)) {return 3;}
	if (tmp_small_page[mi] == 0) {return 4;}
	if (tmp_small_page[mi] == 0xFFFFFFFF) {return 5;}

//	printf("	vpt_test: dbg :i: %d: %X %X %X\n",mi,tmp_small_page[mi],tmp_small_page[mi+4],tmp_small_page[mi+8]); //DBG
	
	return 0; //all OK			
}

/*
 * Function for finding VPT of given process 
 * (resp. PaGeDirectory).
 *
 * This function is assumed to be called only once.
 * When distances used in analyzed images are known,
 * it is faster and safer to use them directly.
 *
 * Function tries address around values that was used in 
 * available linux 2.6 kernel, and tries, if they match definition.
 * Success is not guaranteed, but it was working nice even for kernels
 * with random debugging symbols compiled in.
 *
 * find_process_VPT (linear address of "tasks" list_head)
 * return 0 on success, otherwise error > 0
 * Discovered distances are saved in glibal structure image1.
 *
 * */

int find_process_VPT(unsigned long my_addr){

	int mi=0,tmp_mi;
	unsigned long tmp2,tmp3;

//	printf("find_process_VPT: read: %lx,%d, dst:%d\n",my_addr+TS_MAGIC_HL_MM_DST,POINTER_SIZE,TS_MAGIC_HL_MM_DST);

	for (tmp_mi=0;tmp_mi < TS_MAGIC_HL_MM_DST*2;tmp_mi+=POINTER_SIZE) { //guessing "ts_hl_mm_dst"

		if ( (status=read_linear(0,my_addr+tmp_mi,&tmp2,POINTER_SIZE)) != 0 ) { //first read pointer
	                errormsg=strerror(errno);
	                fprintf(stderr, "find_process_VPT: read() in %s: %s.\n",inputfile,errormsg);
	        } else { //read OK

			for (mi=0;mi < MM_MAGIC_PGD_DST*2;mi+=POINTER_SIZE) { //guessing "mm_pgd_dst"
		
				if ( (status=read_linear(0,tmp2+mi,&tmp3,POINTER_SIZE)) != 0 ) {
		        	        errormsg=strerror(errno);
		               		fprintf(stderr, "find_process_VPT: read() in %s: %s.\n",inputfile,errormsg);
		               		//return 99; //XXX I should handle this in better way XXX
		        	}
				
				if ( user_VPT_test(tmp3) == 0) { //if this pointer points to VPT (resp. PGD)	
					image1.mm_pgd_dst=mi; //then set distance we found (inside mm_struct from start to pgd)
					image1.ts_hl_mm_dst=tmp_mi; // and set distance (inside task_struct from hl "tasks" to mm) 
					printf("Found userspace PGD at %lx and determined distances: mm_pgd: %d, hl_mm: %d\n",tmp3,mi,tmp_mi);
					return 0;
				}
			}
		}

	}
	
	return 1; //did not found VPT(resp. PGD)
}

/*
 * Function that tries to determine general information 
 * about analyzed image.
 *
 * for now, only system VPT is found.
 *
 * determine ()
 * return 0 on success, otherwise error > 0
 * Discovered information are saved in global structure image1.
 *
 * */

int determine(){
	
	int addr=VPT_MAGIC_ADDR;

	//TODO
	
	//Get some ideas, how to do it...
	//Get as much dump as possible
	//Make some observations..
	
	//For now hard-coded: (linux 2.6.23, i386)

	image1.guest_block_size=MB_4; //4M  //size of block in analyzed image 
	//cha-chaa, on real system,we can have 2 block sizes at once!
//        image1.architecture="i386";  // hw architecture of analyzed image
//        image1.operating_system="Linux 2.6.22-3";      // operation system if analyzed image
        image1.ram_size=1048576; //1G   // size of image in kilobytes //!!! fake !!!

// let's find VPT_address:
	
	lseek(inputfd,VPT_MAGIC_ADDR,SEEK_SET); //magic vaule, XXX put to magic section later.
	
	while (VPT_test(addr) != 0) {

		addr += VPT_MAGIC_BS; // XXX magic, 4KB
		lseek(inputfd,addr,SEEK_SET);
		
		if (addr > ( VPT_MAGIC_ADDR + VPT_MAGIC_BS * VPT_MAGIC_COUNT) ) { 
			//we have tried too many pages, so something is wrong probably
			//further search will be useless
			printf("VPT_address not found, terminating at: 0x%X\n",addr);
			return 1;
		}
	}
	
	image1.VPT_address=addr; // found one! :)
	printf("VPT_address found at: 0x%lX\n",image1.VPT_address);
	
	return 0;  // 0 if all OK, otherwise error
}

/* Only printf processes for now.
 * listhead is address of list_head of one of the processes
 * distance is in bytes from given listhead structure to comm (char[16]) field
 */

int list_processes(unsigned long listhead,int distance){
	
	unsigned long tmp_list,tmp_list2;
	unsigned long counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[16];
	int tmp_help=1;

/*	if ( (status=lseek(inputfd,tmp_list_phy,SEEK_SET)) == -1 ) {  
		errormsg = strerror(errno);
	        fprintf(stderr, "list_processes: lseek() to 0x%lx failed because of %s.\n",tmp_list_phy,errormsg);
	        return(-1);
	}
	
	if ( (status=read(inputfd,&tmp_list,POINTER_SIZE)) == -1 ) {
	        errormsg=strerror(errno);
	        fprintf(stderr, "list_processes: read() in %s: %s.\n",inputfile,errormsg);
	        return(-2);
	}
*/	
	printf("Listing processes:\n");
	tmp_list = listhead;

	while (tmp_help) { //This is way "down" in list_head
	
		read_linear(0,tmp_list,&tmp_list2,POINTER_SIZE);
		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {tmp_help=0;} else // we point back to ourselves => circle has closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same task_struct twice. No garanteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					printf("Done.\n");
//					tmp_help=0;
					return counter;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > 65535) { //there cannot be more than 65K processes on Linux (ehm;)
//				tmp_help =0;
				printf("	list_processes: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
			read_linear(0,tmp_list+distance,tmp_comm,16);
			printf("	address: %lx, name: %s\n",tmp_list,tmp_comm); //XXX
		}	
	}

	return -5; //Something is wrong
}

/*
	We do not search for in init directly, because it is too common string in memory => takes too long
	This is almost clone of list_processes
*/

int find_userspace_process(){
	
	unsigned long tmp_list,tmp_list2;
	unsigned long counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[16];
	int tmp_help=1;

	printf("	find_userspace_process: Looking for userspace process %s.\n",LINUX_MAGIC_USER_PROCESS); //I should make this changeable by parameter XXX
	
	tmp_list = image1.ts_one_of_lists; //This is already from list_head, but not userspace

	while (tmp_help) { //This is way "down" in list_head
	
		read_linear(0,tmp_list,&tmp_list2,POINTER_SIZE);
//		printf("	find_userspace_process: tmp:%lx, tmp2:%lx.\n",tmp_list,tmp_list2); 

		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) { //List is empty, or invalid linear addres was provided 
			debug();
			printf("fusp: strange,1: %lx,2: %lx\n",tmp_list,tmp_list2);
			tmp_help=0;
		} else // we point back to ourselves => circle has closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same task_struct twice. No garanteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					printf("	find_userspace_process: Not found %s, something is wrong.\n",LINUX_MAGIC_USER_PROCESS);
//					tmp_help=0;
					return -1;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > 65535) { //there cannot be more than 65K processes on Linux (ehm;)
//				tmp_help =0;
				printf("	find_userspace_process: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
			read_linear(0,tmp_list+image1.ts_hl_comm_dst,tmp_comm,16);
			if ( (strncmp(tmp_comm, LINUX_MAGIC_USER_PROCESS, strlen(LINUX_MAGIC_USER_PROCESS))) == 0 ) {
				printf("Found userspace process %s  at %lx\n",LINUX_MAGIC_USER_PROCESS,tmp_list); //XXX
				image1.ts_one_of_lists=tmp_list;
				return 0;
			}
		}	
	}

	return -5; //Something is wrong, list is empty or invalid
}

/* Another clone of list_processes
 * Only number of members of list_head
 *
 * input: Physical address of list_head
 * output: Number of members, -3 if infinite loop detected
 */

int list_head_size(unsigned long listhead){
	
	unsigned long tmp_list,tmp_list2;
	unsigned long counter=0,loop_detection=0,loop_detection_heuristic=0;
	int tmp_help=1;

	if ( (status=lseek(inputfd,listhead,SEEK_SET)) == -1 ) {  
		errormsg = strerror(errno);
	        fprintf(stderr, "list_head_size: lseek() to 0x%lx failed because of %s.\n",listhead,errormsg);
	        return(-1);
	}
	
	if ( (status=read(inputfd,&tmp_list,POINTER_SIZE)) == -1 ) {
	        errormsg=strerror(errno);
	        fprintf(stderr, "list_head_size: read() in %s: %s.\n",inputfile,errormsg);
	        return(-2);
	}
//	printf("	list_head_size: %lx\n",listhead);

	while (tmp_help) { //This is way "down" in list_head
	
		read_linear(0,tmp_list,&tmp_list2,POINTER_SIZE);
		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {tmp_help=0;} else // we point back to ourselves => circle has closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same task_struct twice. No garanteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
//					printf("	list_head_size: counter:%d.\n",counter);
//					tmp_help=0;
					return counter;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > 65535) { //there cannot be more than 65K processes on Linux (ehm;)
//				tmp_help =0;
				printf("	list_head_size: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
//			read_linear(0,tmp_list+distance,tmp_comm,16);
//			printf("	address: %lx, name: %s\n",tmp_list,tmp_comm); //XXX
		}	
	}

	return -5; //Something is wrong
}

/* Only printf modules for now.
 * listhead is address of list_head of one of the modules (not linear address, but address in physical dump file)
 * distance is in bytes from given listhead structure to comm (char[16]) field
 */


int list_modules(unsigned long listhead,int distance){
	
	unsigned long tmp_list,tmp_list2;
	unsigned long counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[64]; //according to module.h, with  reserve
	int tmp_help=1;

/*	if ( (status=lseek(inputfd,tmp_list_phy,SEEK_SET)) == -1 ) {  
		errormsg = strerror(errno);
	        fprintf(stderr, "list_processes: lseek() to 0x%lx failed because of %s.\n",tmp_list_phy,errormsg);
	        return(-1);
	}
	
	if ( (status=read(inputfd,&tmp_list,POINTER_SIZE)) == -1 ) {
	        errormsg=strerror(errno);
	        fprintf(stderr, "list_processes: read() in %s: %s.\n",inputfile,errormsg);
	        return(-2);
	}
*/	
	printf("Listing modules:\n");
	tmp_list = listhead;

	while (tmp_help) { //This is way "down" in list_head
	
		read_linear(0,tmp_list,&tmp_list2,POINTER_SIZE);
		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {tmp_help=0;} else // we point back to ourselves => circle has closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same module twice. No garanteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					printf("Done.\n");
//					tmp_help=0;
					return counter;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of modules found
			if (loop_detection++ > 65535) { //there cannot be more than 65K modules on Linux (ehm;) XXX chceck XXX
//				tmp_help =0;
				printf("	list_modules: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
			read_linear(0,tmp_list+distance,tmp_comm,64);
			printf("	address: %lx, name: %s\n",tmp_list,tmp_comm); //XXX
		}	
	}

	return -5; //Something is wrong
}

/*
 * Function that tries to determine linux-specific information 
 * about analyzed image.
 *
 * Some process and then some userland process are found,
 * and distance between important fields are determined.
 *
 * Information about "module" structure are determined separately,
 * because module support is only optional.
 *
 * linux_determine ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure image1.
 *
 * */

int linux_determine(){
	
	int tmp;
	unsigned long tmp2=0;
	
	printf("	linux_determine: Looking for process %s\n",LINUX_MAGIC_PROCESS2); //I should make this changeable by parameter XXX

 	if (search_init(inputfile,f,LINUX_MAGIC_PROCESS2) != 0 ) { //This seems to work, even if it is not optimal
		printf("linux_determine: search_init retruned error. Aborting.\n");
		return -1;
	}

	
	while ((tmp2=search_next())!=0){
//		tmp2-=POINTER_SIZE; // because of way how is verify_task_struct written.. maybe I should change this XXX
//		printf("	linux_determine: Trying to verify: %lx as task_struct\n",tmp2);
		if (verify_task_struct(tmp2) == 0) { //we have found valid process
			printf("Found valid process at %lx.\n",tmp2);
			break;
		}
	}

	if (tmp2 == 0) { //Found nothing
		printf("	linux_determine: Cannot find process %s, exiting.",LINUX_MAGIC_PROCESS2); //XXX
		return 2;
	}
//	printf("	linux_determine: tmp2: %lx, image1.ts_one_of_lists: %lx\n",tmp2,image1.ts_one_of_lists);//DBG

	// Find some userspace process
	if ( (tmp=find_userspace_process()) !=0) {
		printf("	linux_determine: find_userspace_process returned :%d, terminating.\n",tmp);
		return 3;
	}
	
//	printf("	linux_determine: 2: tmp2: %lx, image1.ts_one_of_lists: %lx\n",tmp2,image1.ts_one_of_lists);//DBG
	// determnine VPT distances
	if ( (tmp=find_process_VPT(image1.ts_one_of_lists)) !=0) {
		printf("	linux_determine: find_process_VPT returned :%d, terminating.\n",tmp);
		return 4;
	}
      
	printf("Determination process complete.\n");//DBG

	return 0;
}

/*
 * Function that tries to determine linux module-specific 
 * information about analyzed image.
 *
 * Some module is found,
 * and distance between important fields are determined.
 *
 * This is done separately,
 * because module support is only optional, and is not 
 * critical for another analyze.
 *
 * linux_determine_module ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure image1.
 *
 * */

int linux_determine_module(){
	
//	int tmp;
	unsigned long tmp2=0;
	
	printf("	linux_determine_module: Looking for module %s\n",LINUX_MAGIC_MODULE); //I should make this changeable by parameter XXX

 	if (search_init(inputfile,f,"ide_core"/*LINUX_MAGIC_MODULE*/) != 0 ) { //This seems to work, even if it is not optimal
		printf("linux_determine: search_init retruned error. Aborting.\n");
		return -1;
	}

	
	while ((tmp2=search_next())!=0){
//		tmp2-=POINTER_SIZE; // because of way how is verify_task_struct written.. maybe I should change this XXX
//		printf("	linux_determine: Trying to verify: %lx as task_struct\n",tmp2);
//		if (verify_task_struct2(tmp2) == 0) { //we have found valid process
		if (verify_module(tmp2) == 0) { //we have found valid process

			printf("Found valid module at %lx.\n",tmp2);
			break;
		}
	}

	if (tmp2 == 0) { //Found nothing
		printf("	linux_determine: Cannot find module %s, exiting.",LINUX_MAGIC_MODULE); //XXX
		return 2;
	}
//	printf("	linux_determine: tmp2: %lx, image1.ts_one_of_lists: %lx\n",tmp2,image1.ts_one_of_lists);//DBG

	printf("Determination process of module structure complete.\n");//DBG

	return 0;
}


void help(){
	
	printf("\nforiana %s\n",fversion);
	printf("Usage: foriana [options] RAM_image\n");
	printf("Options:\n");
	printf("  -a/--all : try to dump everything avaiable. Generaly not good idea.\n");
	printf("  -c/--cache <file>: specify cache file with already determined information\n");
	printf("  -d/--determine-only : only determine information about image, do not list anything\n");
	printf("  -h/--help : secret option\n");
	printf("  -m/--module <address>  list all information about this module only\n");
	printf("  -o/--output <file>  save output into file\n");
	printf("  -p/--process <address>  list all information about this process only\n");
	printf("  -v/--version : print version number\n");
}

int main(int argc, char **argv)
{	

//	char tempik[8192];

	// get_parameters

	static struct option long_opts[]={
		{"all",0,0,'a'},    
		{"cache",1,0,'c'},    
		{"determine-only",0,0,'d'},    
		{"help",0,0,'h'},
		{"module",1,0,'m'},
//		{"output",1,0,'o'},  
		{"process",1,0,'p'},
//		{"skip-processes",0,0,''},
//		{"skip-modules",0,0,''},
//		{"specify",1,0,'s'},	//specify some options. ex. -s MODULE=loop
		{"version",0,0,'v'},
		{0,0,0,0}                             
	};



	while((opt = getopt_long(argc, argv, "ac:dhm:p:v",long_opts, NULL)) != -1) {
	  switch(opt) {
		case 'a': mode = 2; /*XXX*/ break;
		case 'c': statusfile=optarg; /*XXX*/ break;
		case 'd': mode = 2; /*XXX*/ break;
		case 'h': help(); return 0; break;
		case 'm': mode = 2; /*XXX*/ break;
//		case 'o': outputfile=optarg; break;
		case 'p': mode = 2; /*XXX*/ break;
		case 'v': printf("foriana %s\n",fversion); return 0; break;
		case '?': fprintf(stderr, "Syntax error.\n"); help(); return 7; break;	
	  }   	
	}	

        if ((argv[optind] != NULL) && (argv[optind+1]==NULL)) {
                printf("Analyzing file %s\n",argv[optind]);
		inputfile = argv[optind];
        }else {
		fprintf(stderr, "Syntax error.\n"); 
		help();
		return 7;	
	}

	// parameter defaults
	
	/*
	 * Foriana check for cache file named "." + basename() of analysed image 
	 * inside working directory. If no cache file is found, new cache is
	 * created of this name. Name can be overridden with -c option.
	 *
	 */
	//printf("statusfile0: %s\n",statusfile);fflush(stdout); //DBG
	if ( statusfile == helpstatusfile ) {	// cache file not specified

		if ( (statusfile2=strdup(inputfile) ) == NULL) {	//have to according to man 3 basename
			fprintf(stderr, "malloc() failed.\n");
			return 98;
		}	
		if ( (statusfile=malloc(strlen(inputfile)+1)) == NULL) { //ble...
			fprintf(stderr, "malloc() failed.\n");
			return 99;
		}

		sprintf(statusfile, ".%s", basename(statusfile2));
		free(statusfile2);
		//XXX free statusfile at the end?

	}

	// ----------------- program starts here ---------------------

	//Chceck size of input file & initialize
	if ((i = init()) != 0) {
		printf("Init returned %d. Exiting.\n",i);
		return 1;
	}

	debug();	//XXX DBG
	fflush(stdout);
	if ((i = read_info()) != 0) 
	{
                printf("read_info returned %d. => no cache found.\n",i);
        	
		//Determine maximum information from image
		if ((i = determine()) != 0) {
	                printf("Determine returned %d. Exiting.\n",i);
	                return 2;
	        }
	
		
		if ((i = linux_determine()) != 0) {
                	printf("Linux_determine returned %d. Exiting.\n",i);
                	return 3;
        	}

/*		if ((i = linux_determine_module()) != 0) {
               		printf("Linux_determine returned %d. Exiting.\n",i);
                	return 3;
        	}
*/
		if ((i = save_info()) != 0) {
	                printf("save_info returned %d. Exiting.\n",i);
	                return 5;
	        }


		debug();//DBG
	}
        printf("read_info returned %d\n",i);
	fflush(stdout);

/*        if ((i = determine_open_files(image1.ts_one_of_lists)) != 0) {
              printf("determine_open_files returned %d. Exiting.\n",i);
              return 6;
        }
*/

	//challenge
//	i = verify_task_struct(0x11ad2c30);

/*	//ja2_1g
	i = verify_task_struct(0x18da6d0);
*/
	//ja_1G
//	i = verify_task_struct(0x1ffef194);


//	i = verify_task_struct(0x);


//	printf("Result of verify task_struct: %d\n",i);//	XXX
//	debug();

//	i=list_processes(0xdffef07c,276);
//	printf("2 %lx___________i: %d__________\n",image1.ts_one_of_lists,i);
//	i=list_processes(0xdffc5b60,220-4);
//	printf("3 %lx___________i: %d__________\n",image1.ts_one_of_lists,i);

//	i=list_processes(0xdffef06c,300);//	i=list_processes(0xdffef06c,300);
//	printf("4 %lx___________i: %d__________\n",image1.ts_one_of_lists,i);
	//i=list_processes(0xd1ad2b1c,280);
	//printf("5 %lx___________i: %d__________\n",image1.ts_one_of_lists,i);



//	i=list_processes(image1.ts_one_of_lists,image1.ts_hl_comm_dst);
//	printf("7 %lx___________i: %d__________\n",image1.ts_one_of_lists,i);

//	i=user_VPT_test(0xee5c4000);
//	printf("8 :%d\n",i);

//	i=find_process_VPT(0xf092faf8);	
//	printf("8 :%d\n",i);

//dbg	i=find_process_VPT(0xc18daab8);//init	
//dbg	printf("8 :%d\n",i);

//dbg	i=read_linear(0,0xc18d4ab8,&j,POINTER_SIZE);
//dbg	printf("8 :%d-%lx \n",i,j);

//	printf("main: %s, %s\n",inputfile,"kthreadd\0"); ///XXX       
// 	i=search_init(inputfile,f,"kthreadd\0");       
//	printf("i: %d\n",i);

//	pointer = search_next();
//	printf("pointer: %lx\n",pointer);

/*	if ((i = linux_determine()) != 0) {
                printf("Linux_determine returned %d. Exiting.\n",i);
                return 3;
        }
*/
	if ((i = list_processes(image1.ts_one_of_lists,image1.ts_hl_comm_dst)) != 0) {
                printf("list_processes returned %d.\n");
             //   return 4; //can fail
        }

/*
	if ((i = linux_determine_module()) != 0) {
                printf("Linux_determine returned %d. Exiting.\n",i);
                return 3;
        }

 
	if ((i = list_modules(image1.md_one_of_lists,image1.md_hl_name_dst)) != 0) {
                printf("Linux_determine returned %d.\n");
             //   return 4; //can fail
        }


	debug();
*/
/*			ja_1G (loops!)
	if ((i = list_processes(0xdffef070,image1.ts_hl_comm_dst+4)) != 0) {
                printf("Linux_determine returned %d.\n");
             //   return 4; //can fail
      }
*/

	// ja2_1G
/*	if ((i = list_processes(0xc18da5ac,image1.ts_hl_comm_dst+4)) != 0) {
                printf("Linux_determine returned %d.\n");
             //   return 4; //can fail
      }
*/
	// This is correct for challenge image
/*	if ((i = list_processes(0xd1ad2b20,image1.ts_hl_comm_dst+4)) != 0) {
                printf("Linux_determine returned %d.\n");
             //   return 4; //can fail
        }
*/

//	if ((i = list_processes(0xd1ad2b20,image1.ts_hl_comm_dst+4)) != 0) {
//                printf("list_processes returned %d.\n",i);
             //   return 4; //can fail
//       }

//	if (i != 0 ) { printf("Read linear returned error number %d\n",i); }
	
	printf("Haven't found the lord, oh no no no.\n");
	//close output	
//	close(out_file);	
	return 0;
}
